### Data class and associated helper methods

import numpy as np
import h5py
import os
import time
from threading import Thread
import itertools
import logging

class FilePreloader(Thread):
    def __init__(self, files_list, file_open,n_ahead=2):
        Thread.__init__(self)
        self.deamon = True
        self.n_concurrent = n_ahead
        self.files_list = files_list
        self.file_open = file_open
        self.loaded = {} ## a dict of the loaded objects
        self.should_stop = False
        
    def getFile(self, name):
        ## locks until the file is loaded, then return the handle
        return self.loaded.setdefault(name, self.file_open( name))

    def closeFile(self,name):
        ## close the file and
        if name in self.loaded:
            self.loaded.pop(name).close()
    
    def run(self):
        while not self.files_list:
            time.sleep(1)
        for name in itertools.cycle(self.files_list):
            if self.should_stop:
                break
            n_there = len(self.loaded.keys())
            if n_there< self.n_concurrent:
                logging.debug("preloading {} with {}".format(name,n_there))
                self.getFile( name )
            else:
                time.sleep(5)

    def stop(self):
        logging.debug("Stopping FilePreloader")
        self.should_stop = True

def data_class_getter(name):
    """Returns the specified Data class"""
    data_dict = {
            "H5Data":H5Data,
            }
    try:
        return data_dict[name]
    except KeyError:
        logging.warning("{0:s} is not a known Data class. Returning None...".format(name))
        return None


class Data(object):
    """Class providing an interface to the input training data.
        Derived classes should implement the load_data function.

        Attributes:
          file_names: list of data files to use for training
          batch_size: size of training batches
    """
    def finalize(self):
        if self.caching_directory:            
            for fn in self.relocated:
                logging.debug("removing cached file {}".format(fn))
                os.system('rm -f {}'.format(fn))

    def __init__(self, batch_size, cache=None, copy_command=None):
        """Stores the batch size and the names of the data files to be read.
            Params:
              batch_size: batch size for training
        """
        self.batch_size = batch_size
        self.caching_directory = cache if cache else os.environ.get('DATA_CACHE','')
        self.copy_command = copy_command if copy_command else os.environ.get('DATA_COPY_COMMAND','cp {} {}')
        ## for regular copy it is "cp {} {}"
        ## for s3 it should be "s3cmd get s3://gan-bucket/{} {}"
        ## for xrootd it should be "xrdcp root://cms-xrd-global.cern.ch/{} {}", assuming a valid proxy
        self.fpl = None

    def set_caching_directory(self, cache):
        self.caching_directory = cache
        
    def set_full_file_names(self, file_names):
        self.file_names = list(filter(None, file_names))

    def set_file_names(self, file_names):
        new_file_names = []
        self.relocated = []
        if self.caching_directory:
            goes_to = self.caching_directory
            goes_to += "/"+str(os.getpid())
            os.system('mkdir -p %s '%goes_to)
            for fn in file_names:
                relocate = goes_to+'/'+fn.split('/')[-1]
                if not os.path.isfile( relocate ):
                    logging.debug("Copying {} to {}".format( fn , relocate))
                    cmd = self.copy_command.format( fn, relocate )
                    if os.system(cmd)==0:
                        new_file_names.append( relocate )
                        self.relocated.append( relocate )
                    else:
                        logging.error("was unable to copy the file with {}".format( cmd ))
                        new_file_names.append( fn ) ## use the initial one
                else:
                    new_file_names.append( relocate )
                    self.relocated.append( relocate )
                        
            self.file_names = new_file_names
        else:
            self.file_names = file_names

        if self.fpl:
            self.fpl.files_list = self.file_names

    def inf_generate_data(self):
        while True:
            try:
                for B in self.generate_data():
                    yield B
            except StopIteration:
                logging.warning("start over generator loop")
                
    def generate_data(self):
       """Yields batches of training data until none are left."""
       leftovers = None
       for cur_file_name in self.file_names:
           cur_file_features, cur_file_labels = self.load_data(cur_file_name)
           # concatenate any leftover data from the previous file
           if leftovers is not None:
               cur_file_features = self.concat_data( leftovers[0], cur_file_features )
               cur_file_labels = self.concat_data( leftovers[1], cur_file_labels )
               leftovers = None
           num_in_file = self.get_num_samples( cur_file_features )

           for cur_pos in range(0, num_in_file, self.batch_size):
               next_pos = cur_pos + self.batch_size 
               if next_pos <= num_in_file:
                   yield ( self.get_batch( cur_file_features, cur_pos, next_pos ),
                           self.get_batch( cur_file_labels, cur_pos, next_pos ) )
               else:
                   leftovers = ( self.get_batch( cur_file_features, cur_pos, num_in_file ),
                                 self.get_batch( cur_file_labels, cur_pos, num_in_file ) )

    def count_data(self):
        """Counts the number of data points across all files"""
        num_data = 0
        for cur_file_name in self.file_names:
            cur_file_features, cur_file_labels = self.load_data(cur_file_name)
            num_data += self.get_num_samples( cur_file_features )
        return num_data

    def is_numpy_array(self, data):
        return isinstance( data, np.ndarray )

    def get_batch(self, data, start_pos, end_pos):
        """Input: a numpy array or list of numpy arrays.
            Gets elements between start_pos and end_pos in each array"""
        if self.is_numpy_array(data):
            return data[start_pos:end_pos] 
        else:
            return [ arr[start_pos:end_pos] for arr in data ]

    def concat_data(self, data1, data2):
        """Input: data1 as numpy array or list of numpy arrays.  data2 in the same format.
           Returns: numpy array or list of arrays, in which each array in data1 has been
             concatenated with the corresponding array in data2"""
        if self.is_numpy_array(data1):
            return np.concatenate( (data1, data2) )
        else:
            return [ self.concat_data( d1, d2 ) for d1,d2 in zip(data1,data2) ]

    def get_num_samples(self, data):
        """Input: dataset consisting of a numpy array or list of numpy arrays.
            Output: number of samples in the dataset"""
        if self.is_numpy_array(data):
            return len(data)
        else:
            return len(data[0])

    def load_data(self, in_file):
        """Input: name of file from which the data should be loaded
            Returns: tuple (X,Y) where X and Y are numpy arrays containing features 
                and labels, respectively, for all data in the file

            Not implemented in base class; derived classes should implement this function"""
        raise NotImplementedError

class H5Data(Data):
    """Loads data stored in hdf5 files
        Attributes:
          features_name, labels_name: names of the datasets containing the features
          and labels, respectively
    """
    def __init__(self, batch_size,
                 cache=None,
                 copy_command=None,
                 preloading=0,
                 features_name='features', labels_name='labels'):
        """Initializes and stores names of feature and label datasets"""
        super(H5Data, self).__init__(batch_size,cache,copy_command)
        self.features_name = features_name
        self.labels_name = labels_name
        ## initialize the data-preloader
        self.fpl = None
        if preloading:
            self.fpl = FilePreloader( [] , file_open = lambda n : h5py.File(n,'r'), n_ahead=preloading)
            self.fpl.start()          
       

    def load_data(self, in_file_name):
        """Loads numpy arrays from H5 file.
            If the features/labels groups contain more than one dataset,
            we load them all, alphabetically by key."""
        if self.fpl:
            h5_file = self.fpl.getFile( in_file_name )
        else:
            h5_file = h5py.File( in_file_name, 'r' )
        if type(self.features_name) == tuple:
            ## there is a data adaptor
            feature_name, feature_adaptor = self.features_name
        else:
            feature_adaptor = None
            feature_name = self.features_name
        if type(self.labels_name) == tuple:
            ## there is a data adaptor
            label_name, label_adaptor = self.labels_name
        else:
            label_adaptor = None
            label_name = self.labels_name
        
        X = self.load_hdf5_data( h5_file[feature_name] )
        Y = self.load_hdf5_data( h5_file[label_name] )
        if feature_adaptor is not None:
            X = feature_adaptor(X)
        if label_adaptor is not None:
            Y = label_adaptor(Y)
        if self.fpl:
            self.fpl.closeFile( in_file_name )
        else:
            h5_file.close()
        return X,Y 

    def load_hdf5_data(self, data):
        """Returns a numpy array or (possibly nested) list of numpy arrays 
            corresponding to the group structure of the input HDF5 data.
            If a group has more than one key, we give its datasets alphabetically by key"""
        if hasattr(data, 'keys'):
            out = [ self.load_hdf5_data( data[key] ) for key in sorted(data.keys()) ]
        else:
            out = data[:]
        return out

    def count_data(self):
        """This is faster than using the parent count_data
            because the datasets do not have to be loaded
            as numpy arrays"""
        num_data = 0
        for in_file_name in self.file_names:
            h5_file = h5py.File( in_file_name, 'r' )
            if type(self.features_name) == tuple:
                feature_name = self.features_name[0]
            else:
                feature_name = self.features_name
            X = h5_file[feature_name]
            if hasattr(X, 'keys'):
                num_data += len(X[ X.keys()[0] ])
            else:
                num_data += len(X)
            h5_file.close()
        return num_data

    def finalize(self):
        if self.fpl:
            self.fpl.stop()
        Data.finalize(self)



import torch
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
import torch.nn.functional as F
import numpy
import logging

class MNistNet(nn.Module):
    def __init__(self, **args):
        super(MNistNet, self).__init__()
        ks = int(args.get('kernel_size',5))
        do = float(args.get('dropout',0.5))
        dense = int(args.get('dense',50))
        self.conv1 = nn.Conv2d(1, 10, kernel_size=ks)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=ks)
        self.conv2_drop = nn.Dropout2d(do)
        self.fc1 = nn.Linear(320, dense)
        self.fc2 = nn.Linear(dense, 10)

    def forward(self, x):
        x = x.permute(0,3,1,2).float()
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        #return F.log_softmax(x, dim=1)
        #return F.softmax(x)
        #return F.cross_entropy(x)
        return x



### Build a customized CNN with given hyperparameters
class _ConvBlock(nn.Sequential):
    def __init__(self, conv_layers, dropout, in_ch=5):
        super().__init__()
        for i in range(conv_layers):
            channel_in = in_ch*((i+1)%5)
            channel_out = in_ch*((i+2)%5)
            if channel_in == 0: channel_in += 1
            if channel_out == 0: channel_out += 1
            self.add_module('convlayer%d'%(i), nn.Conv2d(channel_in, out_channels=channel_out,kernel_size=(3,3),stride=1, padding=1))
            self.add_module('relu%d'%(i), nn.ReLU(inplace=True))
        self.add_module('convlayer%d'%(conv_layers), nn.Conv2d(channel_out, out_channels=1, kernel_size=(3,3), stride=2, padding=1))
        self.dropout = dropout

    def forward(self, x):
        x = super().forward(x)
        if self.dropout > 0:
            x = F.dropout(x, p=self.dropout, training=self.training)
        return x

class _DenseBlock(nn.Sequential):
    def __init__(self, dense_layers, dropout ,base):
        super().__init__()
        for i in range(dense_layers):
            il = int(base//(2**i))
            ol = int(base//(2**(i+1)))
            logging.info("{} =>> {}".format(il,ol))
            self.add_module('denselayer%d'%(i), nn.Linear(il, ol))
            self.add_module('relu%d'%(i), nn.ReLU(inplace=True))
        self.dropout = dropout

    def forward(self, x):
        x = super().forward(x)
        if self.dropout > 0:
            x = F.dropout(x, p=self.dropout, training=self.training)
        return x

class CNN(nn.Module):
    def __init__(self, conv_layers=2, dense_layers=2, dropout=0.5, classes=3, in_channels=5):
        super().__init__()
        self.build_net(conv_layers, dense_layers, dropout, classes, in_channels)
    
    def build_net(self,*args, **kwargs):
        base_2 = 10
        base = base_2**2
        self.conv_layers = _ConvBlock(args[0], args[2], args[4])
        self.dense_layers = _DenseBlock(args[1], args[2], base)
        self.adapt_pool = nn.AdaptiveMaxPool2d((base_2,base_2))
        il = int(base//(2**(args[1])))
        ol = int(args[3])
        logging.info("{} =>> {}".format(il,ol))
        self.output = nn.Linear(il, ol)

    def forward(self, x):
        x = x.permute(0,3,1,2).float()        
        x = self.conv_layers(x)
        x = self.adapt_pool(x)
        x = x.view(x.shape[0], -1) # flatten
        x = self.dense_layers(x)
        return self.output(x)
            




import os
import glob
import sys

dest='/bigdata/shared/LCDJets_Abstract_IsoLep_lt_20'
import socket
host = os.environ.get('HOST', os.environ.get('HOSTNAME',socket.gethostname()))
if 'titan' in host:
    dest='/ccs/proj/csc291/DATA/LCDJets_Abstract_IsoLep_lt_20'
train = glob.glob(dest+'/train/*.h5')
test  = glob.glob(dest+'/val/*.h5')

N=10
Nt=N/5
if len(sys.argv)>=1:
    a = sys.argv[1]
    if a.isdigit():
        N = int(a)
        Nt=N/5            
    else:
        N,Nt = map(int, a.split(','))


open('train_topclass.list','w').write( '\n'.join(sorted( train[:N] )))
open('test_topclass.list','w').write( '\n'.join(sorted( test[:Nt] )))



### MPIManager class and associated functions

from __future__ import division
import math
from mpi4py import MPI
import numpy as np
import time
import json
import logging

from ..train.data import H5Data

def get_groups(comm, num_masters=1, num_processes=1):
    masters = list(range(0,num_masters)) # index 0 is the uber master, the other one asre sub-masters
    n_active_master = max(num_masters-1,1)
    groups = [set() for _ in range(n_active_master)]
    instances_ranks = list(range(num_masters, comm.Get_size(), num_processes))
    processes = [None]*len(instances_ranks)
    for ir,i_rank in enumerate(instances_ranks):
        _,ig = divmod( ir, n_active_master )
        groups[ig].add( i_rank )
        groups[ig].add( masters[-1-ig] )
        processes[ir] =  list(range(i_rank,min(i_rank+num_processes, comm.Get_size())))

    groups = [sorted(g) for g in groups]
    return masters, groups, processes
    
def get_device(comm, num_masters=1, gpu_limit=-1, gpu_for_master=False):
    """Arguments:
        comm: MPI intracommunicator containing all processes
        num_masters: number of processes that will be assigned as masters
        gpu_limit: maximum number of gpus to use on one host
        gpu_for_master: whether master processes should be given a gpu
       Returns device name 'cpu' or 'gpuN' appropriate for use with theano""" 
    def get_gpu_list(mem_lim = 2000):
        import gpustat
        stats = gpustat.GPUStatCollection.new_query()
        ids = list(map(lambda gpu: int(gpu.entry['index']), stats))
        ratios = map(lambda gpu: float(gpu.entry['memory.used'])/float(gpu.entry['memory.total']), stats)
        #used = list(map(lambda gpu: float(gpu.entry['memory.used']), stats))
        #unused_gpu = filter(lambda x: x[1] < 100.0, zip(ids, used))
        free = list(map(lambda gpu: float(gpu.entry['memory.total'])-float(gpu.entry['memory.used']), stats))
        unused_gpu = list(filter(lambda x: x[1]  > mem_lim, zip(ids, free)))
        unused_gpu = sorted(unused_gpu, key = lambda o : o[1], reverse=True)
        return [x[0] for x in unused_gpu]

    # Get the ranks of the other processes that share the same host
    # and determine which GPU to take on the host
    if gpu_limit==0:
        logging.info("required to not use gpu")
        dev = 'cpu'
        return dev
    
    rank = comm.Get_rank()    
    host = MPI.Get_processor_name()
    hosts = comm.allgather(host)
    workers_sharing_host = [ i for i in range(comm.Get_size()) if hosts[i] == host ]
    if rank in workers_sharing_host:
        worker_id = workers_sharing_host.index( rank )
    else:
        worker_id = -1
    
    for inode in range( comm.Get_size()):
        if rank == inode:
            gpu_list = get_gpu_list() if gpu_limit else []
            #if gpu_limit>=0:
            #    gpu_list = gpu_list[:gpu_limit] #limit the number of gpu
            if len(gpu_list) == 0:
                logging.info("No free GPU available. Using CPU instead.")
                dev = 'cpu'
            elif worker_id<0:
                ## alone on that machine
                dev = 'gpu%d' % (gpu_list[-1])
                logging.info("Alone on the node {} and taking the last gpu {}".format(host, dev))
            else:
                dev = 'gpu%d' % (gpu_list[worker_id%len(gpu_list)])
                logging.info("Sharing a node {} and taking on the gpu {}".format(host, dev))                
            logging.debug("rank %d can have %s",rank,dev)
        comm.Barrier()
    return dev
            

class MPIManager(object):
    """The MPIManager class defines the topology of the MPI process network
        and creates master and worker objects for each process accordingly.

        Two configurations are available:
          1) one master supervising other masters, each controlling some workers
          2) one master and N-1 workers, where N is the number of MPI processes

        Attributes:
          process: the MPI worker or master object running on this process
          data: Data object containing information used for training/validation
          algo: Algo object containing training algorithm configuration options
          model_builder: ModelBuilder object
          num_masters: integer indicating the number of master processes.  
            If num_masters > 1, an additional master will be created to supervise all masters.
          num_workers: integer indicating the number of worker processes
          worker_id: ID of worker node, used for indexing training data files
          num_epochs (integer): number of times to iterate over the training data
          comm_block: MPI intracommunicator used for message passing between master and workers.
            Process 0 is the master and the other processes are workers.  
          comm_masters: MPI intracommunicator used for message passing between masters.
            (It will be None if there is only one master.)
          train_list: list of training data file names
          val_list: list of validation data file names
          is_master: boolean determining if this process is a master
          should_validate: boolean determining if this process should run training validation
          synchronous: whether or not to syncronize workers after each update
          verbose: whether to make MPIProcess objects verbose
    """

    def __init__(self, comm, data, algo, model_builder, num_epochs, train_list, 
                 val_list, num_masters=1, num_processes=1, synchronous=False,
                 verbose=False, custom_objects={}, early_stopping=None,target_metric=None,
                 monitor=False, thread_validation=False, checkpoint=None, checkpoint_interval=5):
        """Create MPI communicator(s) needed for training, and create worker 
            or master object as appropriate.

            Params: 
            comm: MPI intracommunicator containing all processes
            data: Data object containing information used for training/validation
            algo: Algo object containing training algorithm configuration options
            model_builder: ModelBuilder object
            num_masters: number of master processes
            num_processes: number of processes that make up a worker/master (gpu allreduce)
            num_epochs: number of times to iterate over the training data
            train_list: list of training data files
            val_list: list of validation data files
            synchronous: true if masters should operate in synchronous mode
            verbose: whether to make MPIProcess objects verbose
            monitor: whether to monitor per-process resource (CPU/GPU) usage
        """
        self.data = data
        self.algo = algo
        self.model_builder = model_builder
        self.num_masters = num_masters
        self.num_processes = num_processes
        if comm.Get_size() != 1:
            n_instances,remainder = divmod(comm.Get_size()-self.num_masters, self.num_processes)
        else:
            n_instances,remainder = 1,0
        if remainder:
            logging.info("The accounting is not correct, %d nodes are left behind",remainder)
        self.num_workers = n_instances # - self.num_masters
        self.worker_id = -1

        self.num_epochs = num_epochs
        self.train_list = train_list
        self.val_list = val_list
        self.synchronous = synchronous
        self.verbose = verbose
        self.monitor = monitor
        self.comm_block = None
        self.comm_masters = None
        self.comm_instance = None
        self.is_master = None
        self.should_validate = None
        self.custom_objects=custom_objects
        self.early_stopping = early_stopping
        self.target_metric = target_metric
        self.thread_validation = thread_validation
        self.checkpoint = checkpoint
        self.checkpoint_interval = checkpoint_interval
        self.make_comms(comm)

    def make_comms(self,comm):
        """Define the network topology by creating communicators linking masters with their slaves.
            Set comm_block to contain one master and all of its workers.
            Set comm_masters to contain all masters, including the "super-master" supervising them.
            Set comm_instance to contain the nodes contributing to the woker/master instance
            Define a master or worker object as appropriate for each process.
            If a worker is created, it is assigned some data files to train on.
        """
        # For masters we let child_comm be the communicator used to message the node's 
        # children, and parent_comm be that used to message the node's parents.

        self.parent_rank = 0
        self.is_master = False
        rank = comm.Get_rank()
        masters, groups, processes = get_groups( comm, self.num_masters, self.num_processes)

        logging.debug("masters %s", str(masters))
        if len(masters)>1:
            self.comm_masters = comm.Create( comm.Get_group().Incl( masters ))
        if rank in masters:
            ## make the communicator for masters
            self.is_master = True
        if rank ==0 :
            self.should_validate = True

            
        logging.debug("groups %s", str(groups))
        for igr,gr in enumerate(groups):
            if rank in gr:
                self.comm_block = comm.Split( igr )
                break

        if self.comm_block is None:
            _ = comm.Split( len(groups))
            
        if not self.is_master and self.comm_block:
            self.worker_id = self.comm_block.Get_rank()

        logging.debug("processes %s", str(processes))
        for ipr,pr in enumerate(processes):
            if rank in pr and len(pr)>1:
                ## make the communicator for that process group
                self.comm_instance = comm.Split(ipr)
                break

        if not self.comm_instance:
            _ = comm.Split( len(processes) )
            
        
        if self.comm_instance:
            ids = self.comm_instance.allgather( self.worker_id )
            self.worker_id = list(filter(lambda i:i!=-1, ids))[0]

        logging.debug("Master communicator size: {}".format(self.comm_masters.Get_size() if self.comm_masters else "N/A"))
        logging.debug("Block communicator size: {}".format(self.comm_block.Get_size() if self.comm_block else "N/A"))
        logging.debug("Instance communicator size: {}".format(self.comm_instance.Get_size() if self.comm_instance else "N/A"))
        
        # Case (1)
        if self.num_masters > 1:
            if self.is_master:
                parent_comm = self.comm_masters
                #if rank==0:
                if self.comm_masters.Get_rank() == 0:
                    child_comm = self.comm_masters
                    self.parent_rank = None
                else:
                    child_comm = self.comm_block
        else:
            if self.is_master:
                parent_comm = self.comm_block
                child_comm = self.comm_block
                self.parent_rank = None

        # Process initialization
        if comm.Get_size() != 1:
            from .process import MPIWorker, MPIMaster
            if self.is_master:
                self.set_val_data()
                num_sync_workers = self.get_num_sync_workers(child_comm)
                self.process = MPIMaster( parent_comm, parent_rank=self.parent_rank, 
                                        data=self.data, algo=self.algo, model_builder=self.model_builder, 
                                        child_comm=child_comm, num_epochs=self.num_epochs, 
                                        num_sync_workers=num_sync_workers,
                                        verbose=self.verbose, custom_objects=self.custom_objects,
                                        early_stopping = self.early_stopping, target_metric = self.target_metric,
                                        threaded_validation = self.thread_validation,
                                        checkpoint=self.checkpoint, checkpoint_interval=self.checkpoint_interval
                )
            else:
                self.set_train_data()
                self.process = MPIWorker( data=self.data, algo=self.algo,
                                        model_builder=self.model_builder,
                                        process_comm = self.comm_instance,
                                        parent_comm=self.comm_block,
                                        parent_rank=self.parent_rank, 
                                        num_epochs=self.num_epochs,
                                        verbose=self.verbose,
                                        monitor=self.monitor,
                                        custom_objects=self.custom_objects,
                                        checkpoint=self.checkpoint, checkpoint_interval=self.checkpoint_interval
                )
        else: #Single Process mode
            from .single_process import MPISingleWorker
            self.set_val_data()
            self.set_train_data(use_all=True)
            self.process = MPISingleWorker(data=self.data, algo=self.algo,
                                        model_builder=self.model_builder,
                                        num_epochs=self.num_epochs,
                                        verbose=self.verbose,
                                        monitor=self.monitor,
                                        custom_objects=self.custom_objects,
                                        early_stopping = self.early_stopping,
                                        target_metric = self.target_metric,
                                        checkpoint=self.checkpoint, checkpoint_interval=self.checkpoint_interval)


    def figure_of_merit(self):
        ##if (self.comm_masters and self.comm_masters.Get_rank() == 0) or (self.comm_block.Get_rank() == 0):
        if self.parent_rank is None:
            ## only the uber-master returns a valid fom
            return self.process.model.figure_of_merit()
        else:
            return None

    def train(self):
        if self.parent_rank is None:
            ## start the uber master, as all over masters are self-started
            ## check MPIProcess.__init__ 
            #if self.parent_rank is not None:
            # self.bcast_weights( self.parent_comm )
            # self.train()
            return self.process.train()
        else:
            return None

    def get_num_sync_workers(self, comm):
        """Returns the number of workers the master should wait for
            at each training time step.  Currently set to 95% of the 
            number of workers (or 1 if running asynchronously).  
            comm should be the master's child communicator."""
        if self.synchronous:
            return int( math.ceil( 0.95 * (comm.Get_size() - 1) ) )
        return 1

    def set_train_data(self, use_all=False):
        """Sets the training data files to be used by the current process"""
        logging.debug("number of workers %d",self.num_workers)
        logging.debug("number of files %d",len(self.train_list))
        if use_all:
            files_for_this_worker = self.train_list
        else:
            files_for_this_worker = [fn for  (i,fn) in enumerate(self.train_list) if i%self.num_workers==(self.worker_id-1)]

        logging.debug("Files for worker id{}, rank {}:{}".format(self.worker_id,self.comm_block.Get_rank() if self.comm_block else "N/A",
                                               self.comm_instance.Get_rank() if self.comm_instance else "N/A")
               )
        if not files_for_this_worker:
            logging.error("There are no files for training, this is a fatal issue")
            MPI.COMM_WORLD.Abort()
            
        for f in files_for_this_worker:
            logging.debug("  {0}".format(f))
        self.data.set_file_names( files_for_this_worker )

    def set_val_data(self):
        """Sets the validation data files to be used by the current process
            (only the master process has validation data associated with it)"""
        if not self.should_validate: return None
        logging.debug("Files for validation:" )
        if not self.val_list:
            logging.error("There are no files for validating, this is a fatal issue")
            MPI.COMM_WORLD.Abort()
                                                
        for f in self.val_list:
            logging.debug("  {0}".format(f))
        self.data.set_file_names( self.val_list )

    def close(self):
        self.free_comms()
        self.process.model.close()
    
    def free_comms(self):
        """Free active MPI communicators"""
        if self.process.process_comm is not None:
            logging.debug("holding on %d",self.process.process_comm.Get_size())
            self.process.process_comm.Barrier()
            if self.model_builder.get_backend_name() == 'pytorch':
                import horovod.torch as hvd
            else:
                import horovod.keras as hvd
            logging.debug("Shutting down Horovod")
            hvd.shutdown()
        if self.comm_block is not None:
            self.comm_block.Free()
        if self.comm_masters is not None:
            self.comm_masters.Free()
        if self.comm_instance is not None:
            self.comm_instance.Free()
            

class MPIKFoldManager(MPIManager):
    def __init__( self, NFolds, comm, data, algo, model_builder, num_epochs, train_list, 
                  val_list, num_masters=1,
                  num_process=1,
                  synchronous=False,
                  verbose=False, custom_objects={},
                  early_stopping=None,target_metric=None,
                  monitor=False,
                  checkpoint=None, checkpoint_interval=5):
        self.comm_world = comm
        self.comm_fold = None
        self.fold_num = None

        if NFolds == 1:
            ## make a regular MPIManager
            self.manager = MPIManager(comm, data, algo, model_builder, num_epochs, train_list,
                                      val_list, num_masters,num_process,
                                      synchronous,
                                      verbose, custom_objects,
                                      early_stopping,target_metric,
                                      monitor,
                                      checkpoint=checkpoint, checkpoint_interval=checkpoint_interval)
            return
        
        if int(comm.Get_size() / float(NFolds))<=1:
            logging.warning("There is only one master+one worker per fold, this isn't going to work well")
            #MPI.COMM_WORLD.Abort()
            
        ## actually split further the work in folds
        rank = comm.Get_rank()
        fold_num = int(rank * NFolds / comm.Get_size())
        self.fold_num = fold_num
        self.comm_fold = comm.Split(fold_num)
        logging.debug("For node {}, with block rank {}, send in fold {}".format(MPI.COMM_WORLD.Get_rank(), rank, fold_num))
        self.manager = None

        if val_list:
            logging.warning("MPIKFoldManager would not expect to be given a validation list")
        all_files = train_list+val_list
        from sklearn.model_selection import KFold
        folding = KFold(n_splits = NFolds)
        folds = list(folding.split( all_files ))
        train, test = folds[ fold_num ]
        train_list_on_fold = list(np.asarray(all_files)[ train ])
        val_list_on_fold = list(np.asarray(all_files)[ test ])
        self.manager = MPIManager(self.comm_fold, data, algo, model_builder, num_epochs, train_list_on_fold,
                                  val_list_on_fold, num_masters,num_process,
                                  synchronous,
                                  verbose, custom_objects,
                                  early_stopping,target_metric,monitor,
                                  checkpoint=checkpoint, checkpoint_interval=checkpoint_interval)

    def close(self):
        self.manager.close()

    def train(self):
        self.manager.train()
    
    def figure_of_merit(self):
        fom = self.manager.figure_of_merit()
        if self.comm_fold is not None:
            foms = self.comm_world.allgather( fom )
            # filter out the None values
            foms = list(filter( None, foms))
            ## make the average and rms
            avg_fom = np.mean( foms )
            std_fom = np.std( foms )
            if self.comm_fold.Get_rank()==0:
                logging.info("Figure of merits over {} folds is {}+/-{}".format( len(foms), avg_fom, std_fom))
            return avg_fom
        else:
            if fom is not None:
                logging.info("Figure of merits from single value {}".format( fom ))
            return fom



### ModelBuilder class and associated helper methods

from nnlo.util.utils import load_model
from .optimizer import OptimizerBuilder
import numpy as np
import copy
import sys
import six
import logging

def tell_gpu_memory(label):
    import gpustat
    stats = gpustat.GPUStatCollection.new_query()
    logging.debug("GPU memory usage {0}: {1}".format(label, ",".join(["{0}:{1}".format(gpu.entry['index'], gpu.entry['memory.used']) for gpu in stats])))

def show_torch_memory(label):
    import gc
    import torch
    from functools import reduce
    import operator as op
    logging.info("Memory content for torch {0}".format(label))
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
                logging.info(reduce(op.mul, obj.size()) if len(obj.size()) > 0 else 0, type(obj), obj.size())
        except Exception as e:
            pass

def session(f):
    def wrapper(*args, **kwargs):
        self_obj = args[0]
        if hasattr(self_obj, 'session'):
            with self_obj.graph.as_default():
                with self_obj.session.as_default():
                    return f(*args, **kwargs)
        else:
            return f(*args, **kwargs)
    return wrapper

class MPIModel(object):
    """Class that abstract all details of the model
    """
    def __init__(self, model=None, models=None):
        self.model = model
        self.models = models
        self.histories = {}
        if model and models:
            raise Exception("Cannot specify single and multiple models")

    @session
    def print_metrics(self, metrics):
        if self.model:
            names = self.model.metrics_names
            for name, metric in zip( names, metrics ):
                logging.info("{0}: {1:.3f}".format(name,metric))
        else:
            for im,m in enumerate(self.models):
                names = m.metrics_names
                ametric = metrics[im,...]
                logging.info('model {0} {1}'.format( im ,m.name))
                for name, metric in zip( names,ametric):
                    logging.info("{0}: {1:.3f}".format(name,metric))
                
    @session
    def get_logs(self, metrics, val=False):
        if self.model:
            if val:
                return { 'val_'+name:np.asscalar(metric) for name, metric in
                         zip( self.model.metrics_names, metrics ) }
            else:
                return { name:np.asscalar(metric) for name, metric in
                         zip( self.model.metrics_names, metrics ) }
        else:
            logs = []
            for im,m in enumerate(self.models):
                ametrics = metrics[im,...]
                if val:
                    logs.append({ 'val_'+name:np.asscalar(metric) for name, metric in
                             zip(m.metrics_names, ametrics ) })
                else:
                    logs.append({ name:np.asscalar(metric) for name, metric in
                                  zip(m.metrics_names, ametrics ) })
            return logs
        
    @session
    def update_history(self, items, arg_hist):
        if self.model:
            for m,v in items.items():
                arg_hist.setdefault(m,[]).append(v)
        else:
            for im,(m,it) in enumerate(zip(self.models, items)):
                m_name = "model%s"%im
                try:
                    m_name = m.name
                except:
                    logging.warning("no name attr")
                for m,v in it.items():
                    arg_hist.setdefault(m_name,{}).setdefault(m,[]).append(v)
        self.histories = arg_hist
                       
    @session
    def format_update(self):
        if self.model:
            return [ np.zeros( w.shape, dtype=np.float32 ) for w in self.model.get_weights() ]
        else:
            up = []
            for m in self.models:
                up.append( [ np.zeros( w.shape, dtype=np.float32 ) for w in m.get_weights() ] )
            return up
    
    @session
    def get_weights(self):
        if self.model:
            return self.model.get_weights()
        else:
            l_weights = []
            for m in self.models:
                l_weights.append( m.get_weights() )
            return l_weights
    
    @session
    def set_weights(self, w ):
        if self.model:
            self.model.set_weights( w )
        else:
            for m,mw in zip(self.models, w ):
                m.set_weights( mw )
            
    @session
    def compile(self, **args):
        if 'optimizer' in args and isinstance(args['optimizer'], OptimizerBuilder):
            opt_builder = args['optimizer']
        else:
            opt_builder = None
        if self.model:
            if opt_builder:
                args['optimizer'] = opt_builder.build()
            self.model.compile( **args )
        else:
            for m in self.models:
                if opt_builder:
                    args['optimizer'] = opt_builder.build()
                m.compile( **args )

    @session
    def train_on_batch(self, **args):
        if self.model:
            return np.asarray(self.model.train_on_batch( **args ))
        else:
            h = []
            for m in self.models:
                h.append(m.train_on_batch( **args ))
            return np.asarray(h)
                
    @session
    def test_on_batch(self, **args):
        if self.model:
            return np.asarray(self.model.test_on_batch( **args ))
        else:
            h= []
            for m in self.models:
                h.append(m.test_on_batch( **args ))
            return np.asarray(h)

    @session
    def figure_of_merit(self, **args):
        ## runs like predict trace, and provides a non differentiable figure of merit for hyper-opt
        ## can of course be the validation loss
        if self.model:
            ## return a default value from the validation history
            return (1.-self.histories['val_acc'][-1])
            #return self.histories['val_loss'][-1]
        else: 
            return 0.


    @session
    def save(self, *args,**kwargs):
        if self.model:
            self.model.save( *args, **kwargs )
        else:
            for im,m in enumerate(self.models):
                fn = 'm%d_%s'%( im, args[0])
                m.save( fn, **kwargs )
            
    def close(self):
        if hasattr(self, 'session'):
            self.session.close()
                
class MPITModel(MPIModel):
    """
    adapter of a torch model to fit in the mpi_learn interface
    """
    def __init__(self, model=None, gpus=0):
        MPIModel.__init__(self,model = model)
        self.gpus = gpus
        self.metrics_names = ["loss"]
        self.loss = None
        self.optimizer = None
        self.metrics = []
        self.loss_functions = None

        if self.gpus>0:
            self.model = self.model.cuda()
        if self.gpus >1:
            import torch.nn as nn
            self.model = nn.DataParallel(self.model)
        setattr(self.model, 'metrics_names', self.metrics_names)
        
    def close(self):
        MPIModel.close(self)
        import torch
        #tell_gpu_memory("before deleting model")
        del self.model
        torch.cuda.empty_cache()
        #tell_gpu_memory("after cache release")
        #show_torch_memory("after cache release")

    def format_update(self):
        ws = self.get_weights()
        return [ np.zeros( w.shape, dtype=np.float32 ) for w in ws]
    
    def get_weights(self):
        if self.gpus > 0:
            return copy.deepcopy([i.data.cpu().numpy() for i in list(self.model.parameters())])
        else:
            return copy.deepcopy([i.data.numpy() for i in list(self.model.parameters())])

    def set_weights(self, weights=[]):
        import torch # Don't put it outside because it will break Tensorflow
        for i,weight in enumerate(weights):
            list(self.model.parameters())[i].data.copy_(torch.from_numpy(weight))

    def compile(self, **kwargs):
        import torch.nn
        import torch.optim

        ### need to map the loss string into the relevant torch object
        self.loss = self._build_loss(kwargs['loss'])
        for metric in kwargs['metrics']:
            if metric.lower() == 'acc' or metric.lower() == 'accuracy':
                self.metrics_names.append('acc')
        ## we need a mapping of the kwargs into what is the optimizer that is getting used
        opt_builder = kwargs.get('optimizer')
        if opt_builder is None:
            self.optimizer = torch.optim.SGD(self.model.parameters(), 1.)
        else:
            self.optimizer = opt_builder.build_torch(self.model)

    def _build_loss(self, loss):
        import torch.nn
        if loss and loss.endswith('Loss'):
            loss = loss[:-4]
        lookup = {
            # Keras mapped losses
            'categorical_crossentropy': torch.nn.CrossEntropyLoss,
            'mean_squared_error' : torch.nn.MSELoss,
            'mean_absolute_error' : torch.nn.L1Loss,
            # PyTorch losses 
            'L1': torch.nn.L1Loss,
            'MSE': torch.nn.MSELoss,
            'CrossEntropy': torch.nn.CrossEntropyLoss,
            #'CTC': torch.nn.CTCLoss,
            'NLL': torch.nn.NLLLoss,
            'PoissonNLL': torch.nn.PoissonNLLLoss,
            'KLDiv': torch.nn.KLDivLoss,
            'BCE': torch.nn.BCELoss,
            'BCEWithLogits': torch.nn.BCEWithLogitsLoss,
            'MarginRanking': torch.nn.MarginRankingLoss,
            'HingeEmbedding': torch.nn.HingeEmbeddingLoss,
            'MultiLabelMargin': torch.nn.MultiLabelMarginLoss,
            'SmoothL1': torch.nn.SmoothL1Loss,
            'SoftMargin': torch.nn.SoftMarginLoss,
            'MultiLabelSoftMargin': torch.nn.MultiLabelSoftMarginLoss,
            'CosineEmbedding': torch.nn.CosineEmbeddingLoss,
            'MultiMargin': torch.nn.MultiMarginLoss,
            'TripletMargin': torch.nn.TripletMarginLoss
        }
        if loss in lookup:
            return lookup[loss]()
        else:
            logging.warning("No loss mapping found, using CrossEntropyLoss")
            return torch.nn.CrossEntropyLoss()

    def _accuracy(self, output, target, topk=(1,)):
        """Computes the precision@k for the specified values of k"""
        maxk = max(topk)
        batch_size = target.size(0)
        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))
        
        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(1. / batch_size))
        return res
    def _convert_to_tensor(self, data ):
        import torch
        return torch.from_numpy(data)
        #return torch.as_tensor( data, dtype=torch.DoubleTensor)
        #if hasattr(data, 'keys'):
        #    out = [torch.from_numpy(data[key]) for key in sorted(data.keys())]
        #else:
        #    out = torch.from_numpy(data)
        #return out
        
    def train_on_batch(self, x=None, y=None, *args, **kwargs):
        '''Perform a single gradient update on a single batch of data.
        Attributes:
        x: Pytorch tensor of training data
        y: Pytorch tensor of target data

        Return:
        A list of scalar training loss and a metric specified in the compile method.
        '''
        from torch.autograd import Variable
        x = self._convert_to_tensor(x)
        y = self._convert_to_tensor(y)
        self.model.train()
        self.optimizer.zero_grad()
        target = y.long().max(1)[1] # Pytorch doesn't need 1-hot encoded label. Only the indices of classes.
        #target = y.long()
        if self.gpus>0:
            x = x.cuda()
            target = target.cuda()
        x = Variable(x)
        pred = self.model.forward(x)
        loss = self.loss(pred, Variable(target))
        loss.backward()
        self.optimizer.step()
        l_data = loss.data.numpy() if self.gpus == 0 else loss.data.cpu().numpy()
        self.metrics = [l_data] if l_data.shape==() else [l_data[0]]        
        if 'acc' in self.metrics_names: # compute the accuracy
            acc = self._accuracy(pred.data, target, topk=(1,))[0]
            if self.gpus > 0: acc = acc.cpu()
            self.metrics.append(acc.numpy()[0])
        return np.asarray(self.metrics)


    def test_on_batch(self, x=None, y=None, *args, **kwargs):
        '''Test the model on a single batch of samples. No gradient update is performed.
        Attributes:
        x: Pytorch tensor of test data
        y: Pytorch tensor of target data

        Return:
        A list of scalar training loss and a metric specified in the compile method.
        '''
        from torch.autograd import Variable
        x = self._convert_to_tensor(x)
        y = self._convert_to_tensor(y)        
        self.model.eval()
        target = y.long().max(1)[1] # Pytorch doesn't need 1-hot encoded label. Only the indices of classes.
        #target =y.long()
        if self.gpus > 0:
            x = x.cuda()
            target = target.cuda()
        pred = self.model.forward(Variable(x, volatile=True))
        loss = self.loss(pred, Variable(target, volatile=True))
        l_data = loss.data.numpy() if self.gpus == 0 else loss.data.cpu().numpy()
        self.metrics = [l_data] if l_data.shape==() else [l_data[0]]        
        if 'acc' in self.metrics_names: # compute the accuracy
            acc = self._accuracy(pred.data, target, topk=(1,))[0]
            if self.gpus > 0: acc = acc.cpu()
            self.metrics.append(acc.numpy()[0])
        return np.asarray(self.metrics)

    def save(self, *args,**kwargs):
        import torch
        weights_filename = args[0]+'_w'
        arch_filename = args[0]
        torch.save( self.model.state_dict(), weights_filename)
        torch.save( self.model, arch_filename)
            
class ModelBuilder(object):
    """Class containing instructions for building neural net models.
        Derived classes should implement the build_model and get_backend_name
        functions.

        Attributes:
            comm: MPI communicator containing all running MPI processes
    """

    def __init__(self, comm):
        """Arguments:
            comm: MPI communicator 
        """
        self.comm = comm

    def build_model(self):
        """Should return an uncompiled Keras model."""
        raise NotImplementedError

    def get_backend_name(self):
        """Should return the name of backend framework."""
        raise NotImplementedError

class ModelFromJson(ModelBuilder):
    """ModelBuilder class that builds from model architecture specified
        in a JSON file.
        Attributes:
            filename: path to JSON file specifying model architecture
    """

    def __init__(self, comm, filename=None, custom_objects={}, weights=None):
        self.filename = filename
        self.weights = weights
        self.custom_objects = custom_objects
        super(ModelFromJson, self).__init__(comm)

    def build_model(self, local_session=True):
        if type(self.filename) == list:
            models = []
            for fn in self.filename:
                models.append(load_model(filename=fn))
            return MPIModel(models = models)
        else:        
            return MPIModel(model=load_model(filename=self.filename, custom_objects=self.custom_objects, weights_file=self.weights))

    def get_backend_name(self):
        return 'keras'

class ModelTensorFlow(ModelBuilder):
    """ModelBuilder class that builds from model architecture specified
        in a JSON file. Uses Tensorflow and builds the model on the 
        specified GPU.
        Attributes:
            source: path to JSON file specifying model architecture
                    or Keras model to be cloned
    """

    def __init__(self, comm, source,
            custom_objects={}, weights=None):
        if isinstance(source, six.string_types):
            if source.endswith('.py'):
                module = __import__(source.replace('.py','').replace('/', '.'), fromlist=[None])
                self.model = module.get_model()
                self.filename = None
            else:
                self.filename = source
                self.model = None
        else:
            self.filename = None
            self.model = source
        self.weights = weights
        self.custom_objects = custom_objects
        super(ModelTensorFlow, self).__init__(comm)


    def build_model_aux(self):
        if type(self.filename) == list:
            models = []
            self.weights = self.weights.split(',') if self.weights else [None]*len(self.filename)
            for fn,w in zip(self.filename, self.weights):
                models.append(load_model(filename=fn, weights_file=w))
            return MPIModel(models = models)
        else:
            model = load_model(filename=self.filename, model=self.model,
                               custom_objects=self.custom_objects, weights_file=self.weights)
        return MPIModel(model = model)


    def build_model(self, local_session = True):
        import tensorflow as tf

        if local_session:
            graph = tf.Graph()
            session = tf.compat.v1.Session(graph=graph, config=tf.compat.v1.ConfigProto(
                allow_soft_placement=True, log_device_placement=False,
                gpu_options=tf.compat.v1.GPUOptions(
                        per_process_gpu_memory_fraction=1./self.comm.Get_size()) ) )

            with graph.as_default():
                with session.as_default():
                    ret_model = self.build_model_aux()
                    ret_model.session = session
                    ret_model.graph = graph
                    return ret_model
        else:
            tf.compat.v1.Session( config=tf.compat.v1.ConfigProto(
                allow_soft_placement=True, log_device_placement=False,
                gpu_options=tf.compat.v1.GPUOptions(
                    per_process_gpu_memory_fraction=1./self.comm.Get_size()) ) )
            return self.build_model_aux()

    def get_backend_name(self):
        return 'tensorflow'

class ModelPytorch(ModelBuilder):
    def __init__(self, comm, source,
                 weights = None,
                 gpus=0):
        super(ModelPytorch,self).__init__(comm)
        if isinstance(source, six.string_types):
            if source.endswith('.py'):
                module = __import__(source.replace('.py','').replace('/', '.'), fromlist=[None])
                self.model = module.get_model()
                self.filename = None
            else:
                self.filename = source
                self.model = None
        else:
            self.filename = None
            self.model = source
        self.weights = weights
        self.gpus=gpus

    def build_model(self, local_session=True):
        import torch
        ## free memory used
        #tell_gpu_memory("before building model, before cache release")
        torch.cuda.empty_cache()
        #tell_gpu_memory("before building model, after cache release")
        if self.filename is not None:
            model = torch.load(self.filename)
        elif self.model is not None:
            if True:
                model = copy.deepcopy(self.model)
            else:
                model = self.model
        if self.weights:
            wd = torch.load(self.weights)
            model.load_state_dict(wd)
        return MPITModel(model=model, gpus=self.gpus)

    def get_backend_name(self):
        return 'pytorch'
                                                        



### Algo class
import os
import numpy as np
import logging
from ast import literal_eval
from .optimizer import get_optimizer, MultiOptimizer, OptimizerBuilder

class Algo(object):
    """The Algo class contains all information about the training algorithm.
        Attributes:
          optimizer: instance of the Optimizer class used to compute training updates
          optimizer_name: name of the optimizer
          staleness: difference in time step between master and most recent worker's update
          worker_update_type: whether to send weights or gradients to parent process
          send_before_apply: whether to send weights before applying update 
          step_counter: counts time steps to determine when to sync
            (used for Elastic Averaging SGD)
        See __init__ for list of other supported attributes
          """

    # available options and their default values
    supported_opts = {'loss':'binary_crossentropy',
                      'validate_every':1000,
                      'sync_every':1,
                      'mode':'sgd',
                      'worker_optimizer':'sgd',
                      'worker_optimizer_params':'{}',
                      'elastic_force':None,
                      'elastic_lr':1.0,
                      'elastic_momentum':0,
                      }

    def __init__(self, optimizer, **kwargs):
        """optimizer: string naming an optimization algorithm as defined in Optimizer.get_optimizer()
            Configuration options should be provided as keyword arguments.
            Available arguments are:
               loss: string naming the loss function to be used for training
               validate_every: number of time steps to wait between validations
               sync_every: number of time steps to wait before getting weights from parent
               mode: 'sgd', 'easgd' or 'gem' are supported
               worker_optimizer: string indicating which optimizer the worker should use.
                    (note that if worker_optimizer is sgd and worker_lr is 1, the worker's
                     updates will be the gradients computed at each time step, which is 
                     what is needed for many algorithms.)
               elastic_force: alpha constant in the Elastic Averaging SGD algorithm
               elastic_lr: EASGD learning rate for worker
               elastic_momentum: EASGD momentum for worker
            Optimizer configuration options should be provided as additional
            named arguments (check your chosen optimizer class for details)."""
        for opt in self.supported_opts:
            if opt in kwargs:
                setattr(self, opt, kwargs[opt])
            else:
                setattr(self, opt, self.supported_opts[opt])

        self.optimizer_name = optimizer
        if optimizer is not None:
            optimizer_args = { arg:val for arg,val in kwargs.items() 
                if arg not in self.supported_opts }
            self.optimizer = get_optimizer( optimizer )(**optimizer_args)
        else:
            self.optimizer = None

        """Workers are only responsible for computing the gradient and 
            sending it to the master, so we use ordinary SGD with learning rate 1 and 
            compute the gradient as (old weights - new weights) after each batch."""
        self.worker_optimizer_builder = OptimizerBuilder(self.worker_optimizer, literal_eval(self.worker_optimizer_params))

        self.step_counter = 0
        if self.mode == 'gem':
            self.worker_update_type = 'gem'
        elif self.mode == 'easgd':
            self.worker_update_type = 'weights'
            self.send_before_apply = True
        else:
            self.worker_update_type = 'update'
            self.send_before_apply = False

        # Keep track if internal state was restored
        self.restore = False

    def reset(self):
        ## reset any caching running values
        if self.optimizer:
            self.optimizer.reset()
    
    def get_config(self):
        config = {}
        config['optimizer'] = str(self.optimizer_name)
        for opt in self.supported_opts:
            config[opt] = str(getattr(self, opt))
        return config

    def __str__(self):
        params = [ "{}={!s}".format(opt, getattr(self, opt)) for opt in self.supported_opts ]
        return "Algo(optimizer={}, {})".format(self.optimizer_name, ",".join(params))

    ### For Worker ###

    def compile_model(self, model):
        """Compile the model."""
        model.compile( loss=self.loss, optimizer=self.worker_optimizer_builder, metrics=['accuracy'] )

    def compute_update(self, cur_weights, new_weights):
        """Computes the update to be sent to the parent process"""
        if self.worker_update_type == 'gem':
            return self.optimizer.begin_compute_update(cur_weights, new_weights)
        elif self.worker_update_type == 'weights':
            return new_weights
        else:
            update = []
            for cur_w, new_w in zip( cur_weights, new_weights ):
                if type(cur_w) == list:
                    ## polymorph case
                    update.append([])
                    for sub_c_w,sub_n_w in zip(cur_w, new_w):
                        update[-1].append( np.subtract( sub_c_w,sub_n_w) )
                else:
                    update.append( np.subtract( cur_w, new_w ) )
            return update

    def compute_update_worker(self, weights, update):
        """Compute the update on worker (for GEM)"""
        if self.mode == 'gem': # Only possible in GEM mode
            return self.optimizer.compute_update(weights, update)

    def set_worker_model_weights(self, model, weights):
        """Apply a new set of weights to the worker's copy of the model"""
        if self.mode == 'easgd':
            new_weights = self.get_elastic_update( model.get_weights(), weights )
            model.set_weights( new_weights )
        else:
            model.set_weights( weights )

    def get_elastic_update(self, cur_weights, other_weights):
        """EASGD weights update"""
        new_weights = []
        for m_w,om_w in zip( cur_weights, other_weights ):
            if type( m_w ) == list:
                new_weights.append( [] )
                for cur_w,other_w in zip( m_w, om_w ):
                    new_w = cur_w - self.elastic_force * np.subtract( cur_w, other_w )
                    new_weights[-1].append( new_w )
            else:
                new_w = m_w - self.elastic_force * np.subtract( m_w, om_w )
                new_weights.append( new_w )
        return new_weights

    def should_sync(self):
        """Determine whether to pull weights from the master"""
        self.step_counter += 1
        return self.step_counter % self.sync_every == 0

    ### For Master ###

    def apply_update(self, weights, update):
        """Calls the optimizer to apply an update
            and returns the resulting weights"""
        if self.mode == 'easgd':
            return self.get_elastic_update( weights, update )
        else:
            if type(weights[0]) == list:
                if type(self.optimizer)!= MultiOptimizer:
                    self.optimizer = MultiOptimizer( self.optimizer, len(weights))
            new_weights = self.optimizer.apply_update( weights, update )
            return new_weights
        
    def save(self, fn=None):
        if self.optimizer:
            self.optimizer.save(fn)

    def load(self, fn):
        if not self.optimizer:
            logging.info("No state to restore for this algorithm")
            self.restore = True
            return
        new_optimizer = self.optimizer.load(fn)
        if new_optimizer is not None:
            logging.info("Restored state from {}".format(fn))
            self.optimizer = new_optimizer
            self.restore = True
        else:
            logging.warning("Failed to restore state from {}, starting srom scratch".format(fn))



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='nnlo',
    version='0.0.2',
    description='Distributed Machine Learning tool for High Performance Computing',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='NNLO team',
    author_email='rui.zhang@cern.ch',
    keywords=['Distributed Machine Learning', 'High Performance Computing', 'Hyperparameter optimisation'],
    url='https://github.com/chnzhangrui/NNLO',
    download_url='https://pypi.org/project/nnlo/'
)

install_requires = [
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires, include_package_data=True)



### Predefined Keras models

import sys
import logging
from nnlo.util.utils import import_keras
import_keras()

def model_function(model_name):
    """Constructs the Keras model indicated by model_name"""
    model_maker_dict = {
            'example':make_example_model,
            'mnist':make_mnist_model,
            'cifar10':make_cifar10_model,
            'mnist_torch':make_mnist_torch_model,
            'topclass': make_topclass_model,
            'topclass_torch':make_topclass_torch_model
        
            }
    return model_maker_dict[model_name]    
def make_model(model_name, **args):
    m_fn = model_function(model_name)
    if args and hasattr(m_fn,'parameter_range'):
        provided = set(args.keys())
        accepted = set([a.name for a in m_fn.parameter_range])
        if not provided.issubset( accepted ):
            logging.error("provided arguments {} do not match the accepted ones {}".format(sorted(provided),sorted(accepted)))
            sys.exit(-1)
    return model_function(model_name)(**args)

def make_example_model():
    """Example model from keras documentation"""
    from tensorflow.keras.models import Sequential
    from tensorflow.keras.layers import Dense, Activation
    model = Sequential()
    model.add(Dense(output_dim=64, input_dim=100))
    model.add(Activation("relu"))
    model.add(Dense(output_dim=10))
    model.add(Activation("softmax"))
    return model

def make_topclass_model(**args):
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Dense, Activation, Dropout, Flatten, Input, Permute
    from tensorflow.keras.layers import Convolution2D, MaxPooling2D, Conv2D
    if args:logging.debug("receiving arguments {}".format(args))
    conv_layers=args.get('conv_layers',2)
    dense_layers=args.get('dense_layers',2)
    dropout=args.get('dropout',0.2)
    kernel = args.get('kernel_size',3)
    classes=3
    in_channels=5
    in_ch = in_channels
    ## the trace in the input file is 750, 150, 94, 5
    input = Input( (150,94,in_ch))
    ## convs
    c = input
    for i in range(conv_layers):
        channel_in = in_ch*((i+1)%5)
        channel_out = in_ch*((i+2)%5)
        if channel_in == 0: channel_in += 1
        if channel_out == 0: channel_out += 1
        c = Conv2D( filters=channel_out, kernel_size=(kernel,kernel) , strides=1, padding="same", activation = 'relu') (c)
    c = Conv2D(1, (kernel,kernel), activation = 'relu',strides=2, padding="same")(c)

    ## pooling
    pool = args.get('pool', 10)
    m  = MaxPooling2D((pool,pool))(c)
    f = Flatten()(m)
    d = f
    base = args.get('hidden_factor',5)*100
    for i in range(dense_layers):
        N = int(base//(2**(i+1)))
        d = Dense( N, activation='relu')(d)
        if dropout:
            d = Dropout(dropout)(d)
    o = Dense(classes, activation='softmax')(d)

    model = Model(inputs=input, outputs=o)
    #model.summary()
    return model

def make_cifar10_model(**args):
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Dense, Activation, Dropout, Flatten, Input, Permute
    from tensorflow.keras.layers import Convolution2D, MaxPooling2D, Conv2D
    import tensorflow.keras.backend as K
    if args:logging.debug("receiving arguments {}".format(args))
    nb_classes = 10
    img_rows, img_cols = 32, 32
    
    # use 1 kernel size for all convolutional layers
    ks = args.get('kernel_size', 3)
    
    # tune the number of filters for each convolution layer
    nb_filters1 = args.get('nb_filters1', 48)
    nb_filters2 = args.get('nb_filters2', 96)
    nb_filters3 = args.get('nb_filters3', 192)
    
    # tune the pool size once
    ps = args.get('pool_size', 2)
    pool_size = (ps,ps)
    
    # tune the dropout rates independently
    do4 = args.get('dropout1', 0.25)
    do5 = args.get('dropout2', 0.5)
    
    # tune the dense layers independently
    dense1 = args.get('dense1', 512)
    dense2 = args.get('dense2', 256)
    
    if tensorflow.keras.backend.image_data_format() == 'channels_first':
        input_shape = (3, img_rows, img_cols)
    else:
        input_shape = (img_rows, img_cols, 3)

    #act = 'sigmoid'
    act = 'relu'
        
    i = Input( input_shape)
    l = Conv2D(nb_filters1,( ks, ks), padding='same', activation = act)(i)
    l = MaxPooling2D(pool_size=pool_size)(l)
    #l = Dropout(do1)(l)

    l = Conv2D(nb_filters2, (ks, ks), padding='same',activation=act)(l)
    #l = Conv2D(nb_filters2, (ks, ks))(l)
    l = MaxPooling2D(pool_size=pool_size)(l)
    #l = Dropout(do2)(l)

    l = Conv2D(nb_filters3, (ks, ks), padding='same',activation=act)(l)
    #l = Conv2D(nb_filters3, (ks, ks))(l)
    l = MaxPooling2D(pool_size=pool_size)(l)
    #l = Dropout(do3)(l)

    l = Flatten()(l)
    l = Dense(dense1,activation=act)(l)
    l = Dropout(do4)(l)
    l = Dense(dense2,activation=act)(l)
    l =Dropout(do5)(l)
    
    o = Dense(nb_classes, activation='softmax')(l)

    model = Model(inputs=i, outputs=o)
    #model.summary()
    
    return model

def make_mnist_model(**args):
    import tensorflow
    from tensorflow.keras.models import Sequential, Model
    from tensorflow.keras.layers import Dense, Activation, Dropout, Flatten, Input, Permute
    from tensorflow.keras.layers import Convolution2D, MaxPooling2D, Conv2D
    import tensorflow.keras.backend as K
    """MNIST ConvNet from keras/examples/mnist_cnn.py"""
    #np.random.seed(1337)  # for reproducibility
    if args:logging.debug("receiving arguments {}".format(args))
    nb_classes = 10
    # input image dimensions
    img_rows, img_cols = 28, 28
    # number of convolutional filters to use
    nb_filters = args.get('nb_filters',32)
    # size of pooling area for max pooling
    ps = args.get('pool_size',2)
    
    # convolution kernel size
    ks = args.get('kernel_size',3)
    do = args.get('dropout', 0.25)
    dense = args.get('dense', 128)

    pool_size = (ps,ps)
    if tensorflow.keras.backend.image_data_format() == 'channels_first':
        input_shape = (1, img_rows, img_cols)
    else:
        input_shape = (img_rows, img_cols, 1)
    model = Sequential()
    model.add(Convolution2D(nb_filters, (ks, ks),
                            padding='valid',
                            input_shape=input_shape))
    model.add(Activation('relu'))
    model.add(Convolution2D(nb_filters, (ks, ks)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=pool_size))
    model.add(Dropout(do))
    model.add(Flatten())
    model.add(Dense(dense))
    model.add(Activation('relu'))
    model.add(Dropout(do))
    model.add(Dense(nb_classes))
    model.add(Activation('softmax'))
    return model

def make_mnist_torch_model(**args):
    if args:logging.debug("receiving arguments {}".format(args))
    try:
        from TorchModels import MNistNet
    except:
        from .TorchModels import MNistNet
    model = MNistNet(**args)
    return model

def make_topclass_torch_model(**args):
    if args:logging.debug("receiving arguments {}".format(args))
    conv_layers=args.get('conv_layers',2)
    dense_layers=args.get('dense_layers',2)
    dropout=args.get('dropout',0.5)
    classes=3
    in_channels=5
    try:
        from TorchModels import CNN
    except:
        from .TorchModels import CNN
    model = CNN(conv_layers=conv_layers, dense_layers=dense_layers, dropout=dropout, classes=classes, in_channels=in_channels)
    return model

try:
    from skopt.space import Real, Integer, Categorical
    make_mnist_model.parameter_range =     [
        Integer(10,50, name='nb_filters'),
        Integer(2,10, name='pool_size'),
        Integer(2,10, name='kernel_size'),
        Integer(50,200, name='dense'),
        Real(0.0, 1.0, name='dropout')
    ]
    make_mnist_torch_model.parameter_range = [
        Integer(2,10, name='kernel_size'),
        Integer(50,200, name='dense'),
        Real(0.0, 1.0, name='dropout')
    ]
    make_topclass_model.parameter_range =   [
        Integer(1,6, name='conv_layers'),
        Integer(1,6, name='dense_layers'),
        Integer(1,6, name='kernel_size'),
        Real(0.0, 1.0, name='dropout')
    ]
    make_topclass_torch_model.parameter_range =    [
        Integer(1,6, name='conv_layers'),
        Integer(1,6, name='dense_layers'),
        Real(0.0,1.0, name='dropout')
    ]
    make_cifar10_model.parameter_range = [
        Integer(10,300, name='nb_filters1'),
        Integer(10,300, name='nb_filters2'),
        Integer(10,300, name='nb_filters3'),
        Integer(50,1000, name='dense1'),
        Integer(50,1000, name='dense2'),
        Real(0.0, 1.0, name='dropout1'),
        Real(0.0, 1.0, name='dropout2')
    ]
except:
    pass

